from django import template
from django.utils.html import format_html_join

from django_cotton.templatetags._component import cotton_component
from django_cotton.templatetags._slot import cotton_slot
from django_cotton.templatetags._vars_frame import cotton_vars_frame
from django_cotton.utils import eval_string

register = template.Library()
register.tag("cotton_component", cotton_component)
register.tag("cotton_slot", cotton_slot)
register.tag("cotton_vars_frame", cotton_vars_frame)


@register.filter
def merge(attrs, args):
    # attrs is expected to be a dictionary of existing attributes
    # args is a string of additional attributes to merge, e.g., "class:extra-class"
    for arg in args.split(","):
        key, value = arg.split(":", 1)
        if key in attrs:
            attrs[key] = value + " " + attrs[key]
        else:
            attrs[key] = value
    return format_html_join(" ", '{0}="{1}"', attrs.items())


@register.filter
def eval_default(value, arg):
    return value or eval_string(arg)



import ast

from django import template
from django.template import Node
from django.template.loader import render_to_string


def cotton_component(parser, token):
    """
    Template tag to render a cotton component with dynamic attributes.

    Usage:
        {% cotton_component 'template_path' 'component_key' key1="value1" :key2="dynamic_value" %}
    """
    bits = token.split_contents()
    template_path = bits[1]
    component_key = bits[2]

    kwargs = {}
    for bit in bits[3:]:
        key, value = bit.split("=")
        kwargs[key] = value

    nodelist = parser.parse(("end_cotton_component",))
    parser.delete_first_token()

    return CottonComponentNode(nodelist, template_path, component_key, kwargs)


class CottonComponentNode(Node):
    def __init__(self, nodelist, template_path, component_key, kwargs):
        self.nodelist = nodelist
        self.template_path = template_path
        self.component_key = component_key
        self.kwargs = kwargs

    def render(self, context):
        local_context = context.flatten()
        attrs = {}

        for key, value in self.kwargs.items():
            value = value.strip("'\"")

            if key.startswith(":"):
                key = key[1:]  # Remove ':' prefix
                attrs[key] = self.process_dynamic_attribute(value, context)
            elif value == "":
                attrs[key] = True
            else:
                attrs[key] = value

        # Add the remainder as the default slot
        local_context["slot"] = self.nodelist.render(context)

        # Merge slots and attributes into the local context
        all_slots = context.get("cotton_slots", {})
        component_slots = all_slots.get(self.component_key, {})
        local_context.update(component_slots)
        local_context.update(attrs)

        # We need to check if any dynamic attributes are present in the component slots and move them over to attrs
        if "ctn_template_expression_attrs" in component_slots:
            for expression_attr in component_slots["ctn_template_expression_attrs"]:
                attrs[expression_attr] = component_slots[expression_attr]

        local_context["attrs_dict"] = attrs

        # Reset the component's slots in context to prevent bleeding into sibling components
        if self.component_key in all_slots:
            all_slots[self.component_key] = {}

        context.update({"cotton_slots": all_slots})
        return render_to_string(self.template_path, local_context)

    def process_dynamic_attribute(self, value, context):
        """
        Process a dynamic attribute, resolving template variables and evaluating literals.
        """
        try:
            return template.Variable(value).resolve(context)
        except template.VariableDoesNotExist:
            pass

        # Check for boolean attribute
        if value == "":
            return True

        # Evaluate literal string or pass back raw value
        try:
            return ast.literal_eval(value)
        except (ValueError, SyntaxError):
            return value



from django import template
from django.utils.safestring import mark_safe


def cotton_slot(parser, token):
    """
    Template tag to render a cotton slot.
    is_expression_attr: bool whether the attribute is dynamic or not.
        dynamic attributes are from the use of native tags {{ {% in the component attribute. We put them through as a named
        slot so they can be rendered and provided as a template variable
    """
    try:
        tag_name, slot_name, component_key, *optional = token.split_contents()
        is_expression_attr = optional[0] if optional else None
    except ValueError:
        raise template.TemplateSyntaxError("incomplete c-slot %r" % token.contents)

    nodelist = parser.parse(("end_cotton_slot",))
    parser.delete_first_token()
    return CottonSlotNode(slot_name, nodelist, component_key, is_expression_attr)


class CottonSlotNode(template.Node):
    def __init__(self, slot_name, nodelist, component_key, is_expression_attr):
        self.slot_name = slot_name
        self.nodelist = nodelist
        self.component_key = component_key
        self.is_expression_attr = is_expression_attr

    def render(self, context):
        # Add the rendered content to the context.
        if "cotton_slots" not in context:
            context.update({"cotton_slots": {}})

        output = self.nodelist.render(context)

        # Store the slot data in a component-namespaced dictionary
        if self.component_key not in context["cotton_slots"]:
            context["cotton_slots"][self.component_key] = {}

        context["cotton_slots"][self.component_key][self.slot_name] = mark_safe(output)

        # If the slot is a dynamic attribute, we record it so it can be transferred to attrs in the component
        if self.is_expression_attr:
            if (
                "ctn_template_expression_attrs"
                not in context["cotton_slots"][self.component_key]
            ):
                context["cotton_slots"][self.component_key][
                    "ctn_template_expression_attrs"
                ] = []

            context["cotton_slots"][self.component_key][
                "ctn_template_expression_attrs"
            ].append(self.slot_name)

        return ""



import warnings
import hashlib
import os
import re

from django.template.loaders.base import Loader as BaseLoader
from django.core.exceptions import SuspiciousFileOperation
from django.template import TemplateDoesNotExist
from bs4.formatter import HTMLFormatter
from django.utils._os import safe_join
from django.template import Template
from django.core.cache import cache
from django.template import Origin
from django.conf import settings

from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning

warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


class Loader(BaseLoader):
    is_usable = True

    def __init__(self, engine, dirs=None):
        super().__init__(engine)
        self.cache_handler = CottonTemplateCacheHandler()
        self.template_processor = CottonTemplateProcessor()
        self.dirs = dirs

    def get_contents(self, origin):
        # check if file exists, whilst getting the mtime for cache key
        try:
            mtime = os.path.getmtime(origin.name)
        except FileNotFoundError:
            raise TemplateDoesNotExist(origin)

        cache_key = self.cache_handler.get_cache_key(origin.template_name, mtime)
        cached_content = self.cache_handler.get_cached_template(cache_key)

        if cached_content is not None:
            return cached_content

        template_string = self._get_template_string(origin.name)
        compiled_template = self.template_processor.process(
            template_string, origin.template_name
        )

        self.cache_handler.cache_template(cache_key, compiled_template)

        return compiled_template

    def get_template_from_string(self, template_string):
        """Create and return a Template object from a string. Used primarily for testing."""
        return Template(template_string, engine=self.engine)

    def _get_template_string(self, template_name):
        try:
            with open(template_name, "r") as f:
                return f.read()
        except FileNotFoundError:
            raise TemplateDoesNotExist(template_name)

    def get_dirs(self):
        return self.dirs if self.dirs is not None else self.engine.dirs

    def get_template_sources(self, template_name):
        """Return an Origin object pointing to an absolute path in each directory
        in template_dirs. For security reasons, if a path doesn't lie inside
        one of the template_dirs it is excluded from the result set."""
        for template_dir in self.get_dirs():
            try:
                name = safe_join(template_dir, template_name)
            except SuspiciousFileOperation:
                # The joined path was located outside of this template_dir
                # (it might be inside another one, so this isn't fatal).
                continue

            yield Origin(
                name=name,
                template_name=template_name,
                loader=self,
            )


class UnsortedAttributes(HTMLFormatter):
    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v


class CottonTemplateProcessor:
    DJANGO_SYNTAX_PLACEHOLDER_PREFIX = "__django_syntax__"
    COTTON_VERBATIM_PATTERN = re.compile(
        r"\{% cotton_verbatim %\}(.*?)\{% endcotton_verbatim %\}", re.DOTALL
    )
    DJANGO_TAG_PATTERN = re.compile(r"\{%.*?%\}")
    DJANGO_VAR_PATTERN = re.compile(r"\{\{.*?\}\}")

    def __init__(self):
        self.django_syntax_placeholders = []

    def process(self, content, component_key):
        content = self._replace_syntax_with_placeholders(content)
        content = self._compile_cotton_to_django(content, component_key)
        content = self._replace_placeholders_with_syntax(content)
        return self._revert_bs4_attribute_empty_attribute_fixing(content)

    def _replace_syntax_with_placeholders(self, content):
        """# replace {% ... %} and {{ ... }} with placeholders so they dont get touched
        or encoded by bs4. Store them to later switch them back in after transformation.
        """
        self.django_syntax_placeholders = []

        def replace_pattern(pattern, replacement_func):
            return pattern.sub(replacement_func, content)

        def replace_cotton_verbatim(match):
            inner_content = match.group(1)
            self.django_syntax_placeholders.append(inner_content)
            return f"{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{len(self.django_syntax_placeholders)}__"

        def replace_django_syntax(match):
            self.django_syntax_placeholders.append(match.group(0))
            return f"{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{len(self.django_syntax_placeholders)}__"

        # Replace cotton_verbatim blocks
        content = replace_pattern(self.COTTON_VERBATIM_PATTERN, replace_cotton_verbatim)

        # Replace {% ... %}
        content = replace_pattern(self.DJANGO_TAG_PATTERN, replace_django_syntax)

        # Replace {{ ... }}
        content = replace_pattern(self.DJANGO_VAR_PATTERN, replace_django_syntax)

        return content

    def _compile_cotton_to_django(self, html_content, component_key):
        """Convert cotton <c-* syntax to {%."""
        soup = BeautifulSoup(html_content, "html.parser")

        # TODO: Performance optimisation - Make vars_frame optional, only adding it when the user actually provided
        # vars in a component
        soup = self._wrap_with_cotton_vars_frame(soup)
        self._transform_components(soup, component_key)

        return str(soup.encode(formatter=UnsortedAttributes()).decode("utf-8"))

    def _replace_placeholders_with_syntax(self, content):
        """After modifying the content, replace the placeholders with the django template tags and variables."""
        for i, placeholder in enumerate(self.django_syntax_placeholders, 1):
            content = content.replace(
                f"{self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX}{i}__", placeholder
            )

        return content

    def _revert_bs4_attribute_empty_attribute_fixing(self, contents):
        """
        Removes empty attribute values added by BeautifulSoup to Django template tags.

        BeautifulSoup adds ="" to empty attribute-like parts in HTML-like nodes.
        This method removes these additions for Django template tags.

        Examples:
        - <div {{ something }}=""> becomes <div {{ something }}>
        - <div {% something %}=""> becomes <div {% something %}>
        """
        # Remove ="" after Django variable tags
        contents = contents.replace('}}=""', "}}")

        # Remove ="" after Django template tags
        contents = contents.replace('%}=""', "%}")

        return contents

    def _wrap_with_cotton_vars_frame(self, soup):
        """Wrap content with {% cotton_vars_frame %} to be able to govern vars and attributes. In order to recognise
        vars defined in a component and also have them available in the same component's context, we wrap the entire
        contents in another component: cotton_vars_frame."""
        vars_with_defaults = []
        c_vars = soup.find("c-vars")

        # parse c-vars tag to extract variables and defaults
        if c_vars:
            vars_with_defaults = []
            for var, value in c_vars.attrs.items():
                if value is None:
                    vars_with_defaults.append(f"{var}={var}")
                elif var.startswith(":"):
                    # If ':' is present, the user wants to parse a literal string as the default value,
                    # i.e. "['a', 'b']", "{'a': 'b'}", "True", "False", "None" or "1".
                    var = var[1:]  # Remove the ':' prefix
                    vars_with_defaults.append(f'{var}={var}|eval_default:"{value}"')
                else:
                    # Assuming value is already a string that represents the default value
                    vars_with_defaults.append(f'{var}={var}|default:"{value}"')

            c_vars.decompose()

        # Construct the {% with %} opening tag
        opening = "{% cotton_vars_frame " + " ".join(vars_with_defaults) + " %}"
        closing = "{% endcotton_vars_frame %}"

        # Convert the remaining soup back to a string and wrap it within {% with %} block
        wrapped_content = (
            opening
            + str(soup.encode(formatter=UnsortedAttributes()).decode("utf-8")).strip()
            + closing
        )

        # Since we can't replace the soup object itself, we create new soup instead
        new_soup = BeautifulSoup(wrapped_content, "html.parser")

        return new_soup

    def _transform_components(self, soup, parent_key):
        """Replace <c-[component path]> tags with the {% cotton_component %} template tag"""
        for tag in soup.find_all(re.compile("^c-"), recursive=True):
            if tag.name == "c-slot":
                self._transform_named_slot(tag, parent_key)

                continue

            component_key = tag.name[2:]
            component_path = component_key.replace(".", "/").replace("-", "_")
            opening_tag = f"{{% cotton_component {'cotton/{}.html'.format(component_path)} {component_key} "

            # Store attributes that contain template expressions, they are when we use '{{' or '{%' in the value of an attribute
            expression_attrs = []

            # Build the attributes
            for key, value in tag.attrs.items():
                # BS4 stores class values as a list, so we need to join them back into a string
                if key == "class":
                    value = " ".join(value)

                # check if we have django syntax in the value, if so, we need to extract and place them inside as a slot
                if self.DJANGO_SYNTAX_PLACEHOLDER_PREFIX in value:
                    expression_attrs.append((key, value))
                    continue

                opening_tag += ' {}="{}"'.format(key, value)
            opening_tag += " %}"

            component_tag = opening_tag

            if expression_attrs:
                for key, value in expression_attrs:
                    component_tag += f"{{% cotton_slot {key} {component_key} expression_attr %}}{value}{{% end_cotton_slot %}}"

            if tag.contents:
                tag_soup = BeautifulSoup(tag.decode_contents(), "html.parser")
                self._transform_components(tag_soup, component_key)
                component_tag += str(
                    tag_soup.encode(formatter=UnsortedAttributes()).decode("utf-8")
                )

            component_tag += "{% end_cotton_component %}"

            # Replace the original tag with the compiled django syntax
            new_soup = BeautifulSoup(component_tag, "html.parser")
            tag.replace_with(new_soup)

        return soup

    def _transform_named_slot(self, slot_tag, component_key):
        """Compile <c-slot> to {% cotton_slot %}"""
        slot_name = slot_tag.get("name", "").strip()
        inner_html = "".join(str(content) for content in slot_tag.contents)

        # Check and process any components in the slot content
        slot_soup = BeautifulSoup(inner_html, "html.parser")
        self._transform_components(slot_soup, component_key)

        cotton_slot_tag = f"{{% cotton_slot {slot_name} {component_key} %}}{str(slot_soup.encode(formatter=UnsortedAttributes()).decode('utf-8'))}{{% end_cotton_slot %}}"

        slot_tag.replace_with(BeautifulSoup(cotton_slot_tag, "html.parser"))


class CottonTemplateCacheHandler:
    """Handles caching of cotton templates so the html parsing is only done on first load of each view or component."""

    def __init__(self):
        self.enabled = getattr(settings, "COTTON_TEMPLATE_CACHING_ENABLED", True)

    def get_cache_key(self, template_name, mtime):
        template_hash = hashlib.sha256(template_name.encode()).hexdigest()
        return f"cotton_cache_{template_hash}_{mtime}"

    def get_cached_template(self, cache_key):
        if not self.enabled:
            return None

        return cache.get(cache_key)

    def cache_template(self, cache_key, content, timeout=None):
        if self.enabled:
            cache.set(cache_key, content, timeout=timeout)



from django import template
from django.template.base import token_kwargs
from django.utils.safestring import mark_safe

register = template.Library()


def cotton_vars_frame(parser, token):
    """The job of the vars frame is:
    1. to filter out attributes declared as vars inside {{ attrs }} string.
    2. to provide default values to attributes.
    Because we're effecting variables inside the same component, which is not possible usually, we we wrap
    the vars frame around the contents of the component so we can govern the attributes and vars that are available.
    """
    bits = token.split_contents()[1:]  # Skip the tag name

    tag_kwargs = token_kwargs(bits, parser)

    nodelist = parser.parse(("endcotton_vars_frame",))
    parser.delete_first_token()
    return CottonVarsFrameNode(nodelist, tag_kwargs)


class CottonVarsFrameNode(template.Node):
    def __init__(self, nodelist, kwargs):
        self.nodelist = nodelist
        self.kwargs = kwargs

    def render(self, context):
        # Assume 'attrs' are passed from the parent and are available in the context
        component_attrs = context.get("attrs_dict", {})

        # Initialize vars based on the frame's kwargs and parent attrs
        vars = {}
        for key, value in self.kwargs.items():
            # Check if the var exists in component attrs; if so, use it, otherwise use the resolved default
            if key in component_attrs:
                vars[key] = component_attrs[key]
            else:
                # Attempt to resolve each kwarg value (which may include template variables)
                resolved_value = value.resolve(context)
                vars[key] = resolved_value

        # Overwrite 'attrs' in the local context by excluding keys that are identified as vars
        attrs_without_vars = {k: v for k, v in component_attrs.items() if k not in vars}
        context["attrs_dict"] = attrs_without_vars

        # Provide all of the attrs as a string to pass to the component
        def ensure_quoted(value):
            if isinstance(value, str) and value.startswith('"') and value.endswith('"'):
                return value
            else:
                return f'"{value}"'

        attrs = " ".join(
            [
                f"{key}={ensure_quoted(value)}"
                for key, value in attrs_without_vars.items()
            ]
        )

        context.update({"attrs": mark_safe(attrs)})
        context.update(attrs_without_vars)
        context.update(vars)

        return self.nodelist.render(context)
